// [06/2018] Misha Mikhasenko, mikhail.mikhasenko@gmail.com
// adapted by Vincent Mathieu, mathieuv@indiana.edu

#include <functional>
#include <iostream>
#include <complex>
#include <vector>
#include <random>
#include <map>
#include <iomanip>

#define _USE_MATH_DEFINES
#include <cmath>


#include "Definitions.cc"         // constants

//------------------------------------------------------------------------------//
//------------------------------------------------------------------------------//
//------------------------------------------------------------------------------//

double fourv_prod(const std::vector<double> &p1, const std::vector<double> &p2) {
        if (p1.size() != 4 || p2.size() != 4) std::cerr << "Error: p.size() != 4, something is wrong!\n";
        return p1[3]*p2[3]-p1[0]*p2[0]-p1[1]*p2[1]-p1[2]*p2[2];
}
double inv_masssq(const std::vector<double> &p) {
        return fourv_prod(p,p);
}
double inv_masssq_of_sum(const std::vector<double> &p1, const std::vector<double> &p2) {
        return fourv_prod(p1,p1) + fourv_prod(p2,p2) + 2*fourv_prod(p1,p2);
}

//------------------------------------------------------------------------------//
//----------------------THE EQUATIONS FROM THE APPENDIX B-----------------------//
//-----------------------------------------------------------EPC ?? ??? ??------//
#include "FormalismI.cc"

//------------------------------------------------------------------------------//
//---------------------GENERATOR OF ARTIGICIAL EVENTS---------------------------//
//------------------------------------------------------------------------------//
#include "Generator.cc"


//------------------------------------------------------------------------------//
//------------------------THE PARAMETRIZATION OF ISOBARS------------------------//
//------------------------------------------------------------------------------//

namespace IsobarParams {
  // simple Breit-Wigner
  cd bw(double x, double par[]) {
        double m = par[0];  // mass
        double g = par[1];  // width
        cd amp = 1.0/(m*m - x - cd(0,1.0)*m*g);
        return amp;
  };
  
}

//------------------------------------------------------------------------------//
//------------------THE MAIN: CREATE ISOBARS, CALL DENSITY----------------------//
//------------------------------------------------------------------------------//


int readIso(char chAll[MAXISO], int jAll[MAXISO], int lAll[MAXISO], double parAll[MAXISO][NBPAR]);


int main(int argc, char **argv) {
  
        // read isobar parameter
        char chAll[MAXISO];
        int jAll[MAXISO];
        int lAll[MAXISO];
        double parAll[MAXISO][NBPAR];
        int nbIso = readIso(chAll, jAll, lAll, parAll);
  
  
        double m1sq = FormalismI::m1sq, m2sq = FormalismI::m2sq,
               m3sq = FormalismI::m3sq, m4sq = FormalismI::m4sq;
        // ----------------------------------------
        // create the isobars
        std::vector<FormalismI::isobar> isobars;
        std::vector<cd> couplings(nbIso);
  
        // fill up `isobars`-vector here
        FILE *fileIn = fopen("couplings.txt","r");
        double re, im;
        FormalismI::isobar iso_tmp;
        for (int i = 0; i < nbIso; i++) {
          iso_tmp.ch = chAll[i];
          iso_tmp.j = jAll[i];
          iso_tmp.l = lAll[i];
          for (int j=0; j<NBPAR; j++) {
            iso_tmp.par[j] = parAll[i][j];
          }
          iso_tmp.amp = IsobarParams::bw;
          isobars.push_back(iso_tmp);
          
          fscanf(fileIn, "%lf %lf \n", &re, &im);     // read real and imaginary part of coupling
          couplings[i] = cd(re, im);
        }
        fclose(fileIn);
  
        // ----------------------------------------
        int nb = NBPOINTS;      // number of points in the Dalitz plot
        int nbBins = NBBINS;    // bins in both s and t projections
        // create the bins
        double s_bin[nbBins][2], t_bin[nbBins][2];
        double sdeb = pow(mPi + mK,2), sfin = pow(mB - mPsi,2);
        double tdeb = pow(mPi + mPsi,2), tfin = pow(mB - mK,2);
        double ds = (sfin-sdeb)/(2*nbBins), dt = (tfin-tdeb)/(2*nbBins);
        for (int i=0; i<nbBins ; i++) {
            s_bin[i][0] = sdeb + (2*i+1)*ds;
            t_bin[i][0] = tdeb + (2*i+1)*dt;
            s_bin[i][1] = 0;
            t_bin[i][1] = 0;
        }
  
  
        // ----------------------------------------
        std::random_device rd;
        std::mt19937 en(rd());
        std::uniform_real_distribution<> dist_cosT(-1.0, 1.0);
        std::uniform_real_distribution<> dist_phi(-M_PI, M_PI);

        // create s-distibutrion according to the projected phase space
        auto density = [](double s) -> double {
                               return sqrt(FormalismI::lambda12(s)*FormalismI::lambda34(s))/s;
                       };
        std::vector<double> sv(100);
        sv[0] = pow(sqrt(m3sq)+sqrt(m4sq),2);  sv[sv.size()-1] = pow(sqrt(m2sq)-sqrt(m1sq),2);
        std::vector<double> w(sv.size());
        w[0] = 0; w[w.size()-1] = 0;
        for (uint i = 1; i < sv.size()-1; i++) {
                sv[i] = sv[0] + (sv[sv.size()-1]-sv[0])/(sv.size()-1)*i;
                w[i] = density(sv[i]);
        }
        std::piecewise_linear_distribution<> dist_s(sv.begin(), sv.end(), w.begin());

        // loop over the events
        FILE *fileOut = fopen("Dalitz.txt","w");
        const int Nev = (argc < 2) ? nb : atoi(argv[1]);
        for (int i = 0; i < Nev; i++) {
                double s_input = dist_s(en);
                double cosTh = dist_cosT(en);
                double phi = dist_phi(en);

                std::vector<std::vector<double> > ps = Generator::make_up_some_vectors(s_input, cosTh, phi,
                                                                                       m1sq, m2sq, m3sq, m4sq);
                // this function makes the decay to the muons
                double cosTh_mu = dist_cosT(en);
                double phi_mu = dist_phi(en);
                std::vector<std::vector<double> > qs = Generator::decay_p(ps[0], mMu*mMu, mMu*mMu, cosTh_mu, phi_mu);
                std::vector<std::vector<double> > input_vectors{ps[0],ps[1],ps[2],ps[3],qs[0],qs[1]};
          
                double density = FormalismI::density(isobars, couplings, input_vectors);
                double t_input = inv_masssq_of_sum(ps[0], ps[2]);
                //std::cout << "s, t, density: " << s_input << " " << t_input << " " << density << "\n";
                fprintf(fileOut,"%10.5f  %10.5f %10.5f\n", s_input, t_input, density) ;
          
                // construct the s-projection
                for (int j=0; j<nbBins; j++) {
                  if( fabs(s_input - s_bin[j][0]) < ds ){
                    s_bin[j][1]+= density;
                    continue;
                  }
                }
                // construct the t-projection
                for (int j=0; j<nbBins; j++) {
                  if( fabs(t_input - t_bin[j][0]) < dt ){
                    t_bin[j][1]+= density;
                    continue;
                  }
                }

          
        }
        fclose(fileOut);
  
        FILE *fileS = fopen("sproj.txt","w");
        FILE *fileT = fopen("tproj.txt","w");
        // write the projections
        for (int j=0; j<nbBins; j++) {
          fprintf(fileS,"%10.5f  %10.5f\n", s_bin[j][0], s_bin[j][1]) ;
          fprintf(fileT,"%10.5f  %10.5f\n", t_bin[j][0], t_bin[j][1]) ;
        }
        fclose(fileS); fclose(fileT);
  
  
  
        return 0.0;
}



int readIso(char chAll[MAXISO], int jAll[MAXISO], int lAll[MAXISO], double parAll[MAXISO][NBPAR]){
  // read the isobar parameters from input.txt
  
  int nbIso;           // number of isobars
  
  FILE *fileIn = fopen("input.txt","r");
  fscanf(fileIn, "%i\n", &nbIso);
  
  for (int i=0; i<nbIso; i++) {
    fscanf(fileIn, "%c %i %i %lf %lf \n", &chAll[i], &jAll[i], &lAll[i], &parAll[i][0], &parAll[i][1]);
  }
  fclose(fileIn);
  
  return nbIso;
}

